@@ -10,8 +10,7 @@ import logging
 import time
 import sys
 import threading
-import tornado.concurrent
-from tornado.gen import coroutine, Return
+from tornado import concurrent, gen
 import tornado.stack_context
 import uuid
 import yaml
@@ -73,10 +72,10 @@ yaml.representer.Representer.add_representer(
     ProcessStateMachineMeta, yaml.representer.Representer.represent_name)
 
 
-@persistence.auto_persist('_pid', '_CREATION_TIME', '_future', '_paused_flag')
+@persistence.auto_persist('_pid', '_CREATION_TIME', '_future', '_status', '_pre_paused_status')
 class Process(
-        with_metaclass(ProcessStateMachineMeta, StateMachine,
-                       persistence.Savable)):
+    with_metaclass(ProcessStateMachineMeta, StateMachine,
+                   persistence.Savable)):
     """
     The Process class is the base for any unit of work in plumpy.
 
@@ -118,11 +117,12 @@ class Process(
     _spec_type = ProcessSpec
     # Default placeholders, will be populated in init()
     _waiting_callbacks = None
-    _paused_callbacks = None
     _terminated_callbacks = None
     _stepping = False
-    _pausing = None
+    _pausing = None  # type: futures.Future
+    _paused = None
     _killing = None
+    _interrupt_action = None
 
     @classmethod
     def current(cls):
@@ -237,8 +237,9 @@ class Process(
 
         self._setup_event_hooks()
 
-        self._paused_flag = False
-        self._pausing = None  # If pausing, this will be a future
+        self._status = None  # May hold a current status message
+        self._pre_paused_status = None  # Save status when a pause message replaces it, such that it can be restored
+        self._paused = None
 
         # Input/output
         self._raw_inputs = None if inputs is None else utils.AttributesFrozendict(
@@ -259,7 +260,6 @@ class Process(
     def init(self):
         """ Any common initialisation stuff after create or load goes here """
         self._waiting_callbacks = []
-        self._paused_callbacks = []
         self._terminated_callbacks = []
 
         if self._communicator is not None:
@@ -336,26 +336,15 @@ class Process(
             return _LOGGER
 
     @property
-    def paused(self):
-        return self._paused
+    def status(self):
+        return self._status
+
+    def set_status(self, status):
+        self._status = status
 
     @property
-    def _paused(self):
-        return self._paused_flag
-
-    @_paused.setter
-    def _paused(self, paused):
-        if self._paused == paused:
-            return
-
-        self._pausing = None
-
-        # We are changing the paused state
-        self._paused_flag = paused
-        if self._paused_flag:
-            call_with_super_check(self.on_paused)
-        else:
-            call_with_super_check(self.on_playing)
+    def paused(self):
+        return self._paused is not None
 
     def future(self):
         return self._future
@@ -400,8 +389,7 @@ class Process(
         try:
             return self._state.successful
         except AttributeError:
-            raise exceptions.InvalidStateError(
-                'process is not in the finished state')
+            raise exceptions.InvalidStateError('process is not in the finished state')
 
     def killed(self):
         return self.state == process_states.ProcessState.KILLED
@@ -410,7 +398,7 @@ class Process(
         if isinstance(self._state, process_states.Killed):
             return self._state.msg
         else:
-            raise exceptions.InvalidStateError("Has not been killed")
+            raise exceptions.InvalidStateError('Has not been killed')
 
     def exception(self):
         if isinstance(self._state, process_states.Excepted):
@@ -438,7 +426,7 @@ class Process(
         (this needn't be a method).  If it raises an exception it will cause
         the process to fail.
         """
-        args = (callback, ) + args
+        args = (callback,) + args
         handle = events.ProcessCallback(self, self._run_task, args, kwargs)
         self._loop.add_callback(handle.run)
         return handle
@@ -464,19 +452,13 @@ class Process(
         _process_stack().append(self)
         try:
             yield
-        except Exception:
-            if not self._state.is_terminal():
-                self.transition_to(process_states.ProcessState.EXCEPTED,
-                                   *sys.exc_info()[1:])
-            else:
-                raise
         finally:
             assert Process.current() is self, \
                 "Somehow, the process at the top of the stack is not me, " \
                 "but another process! ({} != {})".format(self, Process.current())
             _process_stack().pop()
 
-    @coroutine
+    @gen.coroutine
     def _run_task(self, fn, *args, **kwargs):
         """
         This method should be used to run all Process related functions and coroutines.
@@ -492,7 +474,7 @@ class Process(
         result = yield tornado.stack_context.run_with_stack_context(
             tornado.stack_context.StackContext(self._process_scope),
             functools.partial(coro, *args, **kwargs))
-        raise Return(result)
+        raise gen.Return(result)
 
     # endregion
 
@@ -541,7 +523,6 @@ class Process(
         else:
             self._loop = events.get_event_loop()
 
-        self._pausing = None  # If pausing, this will be a future
         self._state = self.recreate_state(saved_state['_state'])
 
         if 'communicator' in load_context:
@@ -600,12 +581,6 @@ class Process(
     def remove_on_waiting_callback(self, callback):
         self._waiting_callbacks.remove(callback)
 
-    def add_on_paused_callback(self, callback):
-        self._paused_callbacks.append(callback)
-
-    def remove_on_paused_callback(self, callback):
-        self._paused_callbacks.remove(callback)
-
     def add_on_terminated_callback(self, callback):
         self._terminated_callbacks.append(callback)
 
@@ -654,7 +629,7 @@ class Process(
             except ConnectionClosed:
                 self.logger.info(
                     'no connection available to broadcast state change from {} to {}'.
-                    format(from_label, self.state.value))
+                        format(from_label, self.state.value))
 
     def on_exiting(self):
         state = self.state
@@ -717,15 +692,35 @@ class Process(
             cb(self)
 
     @super_check
-    def on_paused(self):
+    def on_pausing(self, msg=None):
+        """ The process is being paused """
+        pass
+
+    @super_check
+    def on_paused(self, msg=None):
         """ The process was paused """
+        self._pausing = None
+
+        # Create a future to represent the duration of the paused state
+        self._paused = futures.Future()
+
+        # Save the current status and potentially overwrite it with the passed message
+        self._pre_paused_status = self.status
+        if msg is not None:
+            self.set_status(msg)
+
         self._fire_event(ProcessListener.on_process_paused)
-        for cb in self._paused_callbacks:
-            cb(self)
 
     @super_check
     def on_playing(self):
         """ The process was played """
+        # Done being paused
+        self._paused.set_result(True)
+        self._paused = None
+
+        self.set_status(self._pre_paused_status)
+        self._pre_paused_status = None
+
         self._fire_event(ProcessListener.on_process_played)
 
     @super_check
@@ -757,10 +752,12 @@ class Process(
 
     @super_check
     def on_kill(self, msg):
+        self.set_status(msg)
         self.future().set_exception(exceptions.KilledError(msg))
 
     @super_check
     def on_killed(self):
+        self._killing = None
         self._fire_event(ProcessListener.on_process_killed, self.killed_msg())
 
     def on_terminated(self):
@@ -780,7 +777,7 @@ class Process(
         if intent == process_comms.Intent.PLAY:
             result = self.play()
         elif intent == process_comms.Intent.PAUSE:
-            result = self.pause()
+            result = self.pause(msg=msg.get(process_comms.MESSAGE_KEY, None))
         elif intent == process_comms.Intent.KILL:
             result = self.kill(msg=msg.get(process_comms.MESSAGE_KEY, None))
         elif intent == process_comms.Intent.STATUS:
@@ -813,29 +810,87 @@ class Process(
             self.transition_to(process_states.ProcessState.EXCEPTED, exception,
                                trace)
 
-    def pause(self):
+    def pause(self, msg=None):
         """
         Pause the process.  Returns True if after this call the process is paused, False otherwise
 
+        :param msg: an optional message to set as the status. The current status will be saved in the private
+            `_pre_paused_status attribute`, such that it can be restored when the process is played again.
         :return: True paused, False otherwise
         """
-        if self._paused:
+        if self.has_terminated():
+            return False
+
+        if self.paused:
             # Already paused
             return True
 
-        if self._pausing:
+        if self._pausing is not None:
             # Already pausing
             return self._pausing
 
         if self._stepping:
             # Ask the step function to pause by setting this flag and giving the
             # caller back a future
-            self._pausing = futures.Future()
-            self._state.interrupt(process_states.Interrupt.PAUSE)
-            return self._pausing
+            interrupt_exception = process_states.PauseInterruption(msg)
+            self._set_interrupt_action_from_exception(interrupt_exception)
+            self._pausing = self._interrupt_action
+            # Try to interrupt the state
+            self._state.interrupt(interrupt_exception)
+            return self._interrupt_action
         else:
-            self._paused = True
-            return True
+            return self._do_pause(msg)
+
+    def _do_pause(self, state_msg, next_state=None):
+        """ Carry out the pause procedure, optionally transitioning to the next state first"""
+        try:
+            if next_state is not None:
+                self.transition_to(next_state)
+            call_with_super_check(self.on_pausing, state_msg)
+            call_with_super_check(self.on_paused, state_msg)
+        finally:
+            self._pausing = None
+
+        return True
+
+    def _create_interrupt_action(self, exception):
+        """
+        Create an interrupt action from the corresponding interrupt exception
+        :param exception: The interrupt exception
+        :type exception: :class:`plumpy.InterruptException`
+        :return: The interrupt action
+        :rtype: :class:`plumpy.CancellableAction`
+        """
+        if isinstance(exception, process_states.PauseInterruption):
+            do_pause = functools.partial(self._do_pause, str(exception))
+            return futures.CancellableAction(do_pause, cookie=exception)
+
+        elif isinstance(exception, process_states.KillInterruption):
+            def do_kill(_next_state):
+                try:
+                    # Ignore the next state
+                    self.transition_to(process_states.ProcessState.KILLED, str(exception))
+                    return True
+                finally:
+                    self._killing = None
+
+            return futures.CancellableAction(do_kill, cookie=exception)
+        else:
+            raise ValueError("Got unknown interruption type '{}'".format(type(exception)))
+
+    def _set_interrupt_action(self, new_action):
+        """
+        Set the interrupt action cancelling the current one if it exists
+        :param new_action: The new interrupt action to set
+        """
+        if self._interrupt_action is not None:
+            self._interrupt_action.cancel()
+        self._interrupt_action = new_action
+
+    def _set_interrupt_action_from_exception(self, interrupt_exception):
+        """ Set an interrupt action from the corresponding interrupt exception """
+        action = self._create_interrupt_action(interrupt_exception)
+        self._set_interrupt_action(action)
 
     def play(self):
         """
@@ -843,14 +898,15 @@ class Process(
 
         :return: True if playing, False otherwise
         """
-        if not self._paused:
-            if self._pausing:
+        if not self.paused:
+            if self._pausing is not None:
                 # Not going to pause after all
-                self._pausing.set_result(False)
+                self._pausing.cancel()
                 self._pausing = None
+                self._set_interrupt_action(None)
             return True
 
-        self._paused = False
+        call_with_super_check(self.on_playing)
         return True
 
     @event(from_states=(process_states.Running, process_states.Waiting))
@@ -891,9 +947,11 @@ class Process(
         if self._stepping:
             # Ask the step function to pause by setting this flag and giving the
             # caller back a future
-            self._killing = futures.Future()
-            self._state.interrupt(process_states.Interrupt.KILL)
-            return self._killing
+            interrupt_exception = process_states.KillInterruption(msg)
+            self._set_interrupt_action_from_exception(interrupt_exception)
+            self._killing = self._interrupt_action
+            self._state.interrupt(interrupt_exception)
+            return self._interrupt_action
         else:
             self.transition_to(process_states.ProcessState.KILLED, msg)
             return True
@@ -929,57 +987,54 @@ class Process(
         :return: None if not terminated, otherwise `self.outputs`
         """
         if not self.has_terminated():
-            if self._paused:
-                self.play()
+            self.loop().run_sync(self.step_until_terminated)
 
-            loop = self._loop
-            loop.run_sync(self.step_until_terminated)
+        return self.future().result()
 
-        if self.has_terminated():
-            self.result()
-            return self.outputs
-
-    @coroutine
+    @gen.coroutine
     def step(self):
         assert not self.has_terminated(), "Cannot step, already terminated"
 
-        if self._paused:
-            return
+        if self.paused:
+            yield self._paused
 
         try:
             self._stepping = True
-            interrupted = False
+            next_state = None
             try:
                 next_state = yield self._run_task(self._state.execute)
-            except process_states.KillInterruption:
-                assert self._killing or self._pausing
-                interrupted = True
+            except process_states.Interruption as exception:
+                # If the interruption was caused by a call to a Process method then there should
+                # be an interrupt action ready to be executed, so just check if the cookie matches
+                # that of the exception i.e. if it is the _same_ interruption.  If not cancel and
+                # build the interrupt action below
+                if self._interrupt_action is not None:
+                    if self._interrupt_action.cookie is not exception:
+                        self._set_interrupt_action_from_exception(exception)
+                else:
+                    self._set_interrupt_action_from_exception(exception)
 
-            if self._killing:
-                self.transition_to(process_states.ProcessState.KILLED, None)
-                self._killing.set_result(True)
-                self._killing = None
+            except KeyboardInterrupt:
+                raise
+            except Exception:
+                # Overwrite the next state to go to excepted directly
+                exc_info = sys.exc_info()
+                next_state = self.create_state(process_states.ProcessState.EXCEPTED, exc_info[1], exc_info[2])
+                self._set_interrupt_action(None)
 
-                # If we were also pausing, then kill takes precedence and
-                # the pause is abandoned
-                if self._pausing:
-                    self._pausing.set_result(False)
-                    self._pausing = None
-            elif not interrupted:
-                # If it wasn't interrupted and we're not killing the process
-                # then transition to the next state and below we deal with the
-                # case that the process may have been externally paused
+            if self._interrupt_action:
+                self._interrupt_action.run(next_state)
+            else:
+                # Everything nominal so transition to the next state
                 self.transition_to(next_state)
 
-            if self._pausing:
-                self._pausing.set_result(True)
-                self._paused = True
         finally:
             self._stepping = False
+            self._set_interrupt_action(None)
 
-    @coroutine
+    @gen.coroutine
     def step_until_terminated(self):
-        while not self.has_terminated() and not self._paused:
+        while not self.has_terminated():
             yield self.step()
 
     # endregion
@@ -1045,7 +1100,7 @@ class Process(
                 elif port.required:
                     raise ValueError(
                         'Value not supplied for required inputs port {}'.
-                        format(name))
+                            format(name))
                 else:
                     continue
             else:
